local SLIPSTREAMINTEROP_VERSION = 1

-- avoid redefiniton on updates
if timetravel.SLIPSTREAMINTEROP_VERSION == nil or timetravel.SLIPSTREAMINTEROP_VERSION < SLIPSTREAMINTEROP_VERSION then

local cv_slipstream = nil
local cv_t3chargedist = nil
local cv_t2chargedist = nil
local cv_maxdistance = nil
local cv_minimumspeed = nil
local cv_maxangle = nil
local cv_chargetoboost = nil
local cv_boosttime = nil
local cv_mindraftspeedboostpercent = nil
local cv_maxdraftspeedboostpercent = nil

local function Rescale(value, oldmin, oldmax, newmin, newmax)
	return newmin + FixedMul(FixedDiv( value-oldmin , oldmax-oldmin), newmax-newmin)
end

local function Clamp(value, newmin ,newmax)
	return min(newmax,max(newmin,value))
end

local fast
local function SpawnFastLines(mo, scale, color)
	fast = P_SpawnMobj(mo.x + (P_RandomRange(-36,36) * mo.scale),
					   mo.y + (P_RandomRange(-36,36) * mo.scale),
					   mo.z + (mo.height/2) + (P_RandomRange(-20,20) * mo.scale),
					   MT_FASTLINE)
	fast.angle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	fast.momx = 3*mo.momx/4
	fast.momy = 3*mo.momy/4
	fast.momz = 3*mo.momz/4
	fast.color = color
	fast.colorized = true
	fast.scale = FixedMul(scale,mo.scale)
	K_MatchGenericExtraFlags(fast, mo)
end

local seekangle
local T3_DISTANCE,T2_DISTANCE,MAX_DISTANCE
local clampedkartspeed,skincolor,slipstreamboost,slipstreamaccel,slipstreamtarget,angletotarget,disttotarget,seekangle,charge
local p,pks,tmo

local function getPlayerInSlipstreamRange(mo, trackingEchoes)

	if (P_IsObjectOnGround(mo)
	and FixedDiv(p.speed, mapobjectscale)>>FRACBITS >= cv_minimumspeed.value -- must be moving decently on the ground
	) then
		
		for potentialtarget in players.iterate do          
			tmo = potentialtarget.mo
			
			if (tmo and tmo.valid -- must have valid player mapobject
			and tmo ~= mo -- Can't splipstream off yourself
			and not potentialtarget.kartstuff[k_hyudorotimer] -- or ghosts
			and P_IsObjectOnGround(tmo) -- or airbourne karts
			and FixedDiv(potentialtarget.speed, mapobjectscale)>>FRACBITS >= cv_minimumspeed.value -- or slowpokes
			) then
				if trackingEchoes and tmo.timetravel and tmo.linkedItem and tmo.linkedItem.valid then
					tmo = tmo.linkedItem
				else continue
				end
				
				-- Calculating the angle at which we look for suitable draft targets based on drifting
				seekangle = mo.angle-(131235115)*pks[k_drift] -- The magic number is 11 degrees, I'm optimizing this way >_>
				angletotarget = abs(seekangle - R_PointToAngle2(mo.x, mo.y, tmo.x, tmo.y))
				disttotarget = FixedHypot(mo.x - tmo.x, mo.y - tmo.y)
				
				if pks[k_drift] then
					if angletotarget > FixedAngle(cv_maxangle.value*2) then continue end
					if disttotarget > 2*MAX_DISTANCE/5 then continue end -- max slipstream distance while drifting
				else 
					if angletotarget > FixedAngle(cv_maxangle.value) then continue end
					if disttotarget > MAX_DISTANCE then continue end -- max slipstream distance
				end -- Narrow angle for following to slipstream

				-- print(p.mo.skin + " " + potentialtarget.mo.skin + " angletotarget: " + angletotarget + " disttotarget: " + disttotarget + " slipstreamtarget: " + p.slipstreamtarget)

				slipstreamtarget = potentialtarget
				return true
			end
		end
	end
	
	return false

end

addHook("MobjThinker", function(mo)
	if timetravel.SLIPSTREAMINTEROP_VERSION > SLIPSTREAMINTEROP_VERSION then return end
	if not timetravel.isActive then return end
	if not (cv_slipstream and cv_slipstream.value and BLib) then	return end -- Has to be turned on

	p = mo.player
	pks = mo.player.kartstuff
	
	if mo.state == S_KART_SPIN or pks[k_squishedtimer] then -- no slipstreaming if you've bumped or spun out
		return
	end

    T3_DISTANCE = cv_t3chargedist.value * mapobjectscale
    T2_DISTANCE = cv_t2chargedist.value * mapobjectscale
    MAX_DISTANCE = cv_maxdistance.value * mapobjectscale

	-- reset until we find a valid slipstreamtarget this frame
	slipstreamboost = 0
	slipstreamaccel = 0
	slipstreamtarget = nil

	if getPlayerInSlipstreamRange(mo, false) then return end
	getPlayerInSlipstreamRange(mo, true)

	-- add/remove slipstream charge
	if slipstreamtarget then
		if disttotarget > T2_DISTANCE then
			charge = 1
		elseif disttotarget > T3_DISTANCE then
			charge = 2
		else
			charge = 2
		end

		-- Juggy: +1 here to counteract BLib's Slipstream natural decrease.
		mo.slipstreamcharge = min($+charge+1, cv_chargetoboost.value)

		if mo.slipstreamcharge >= cv_chargetoboost.value then
			mo.slipstreamtimer = cv_boosttime.value
			if mo.slipstreamsoundtimer == false then
				S_StartSoundAtVolume(mo, sfx_s3k82, INT32_MAX)
				mo.slipstreamsoundtimer = true
			end
		end

		-- also spawn mini lines if charging to teach/show player the charging area
		if mo.slipstreamtimer == 0 then
			SpawnFastLines(mo, FRACUNIT/2, SKINCOLOR_WHITE)

			clampedkartspeed = Clamp(p.kartspeed, 1, 9)
			
			-- Set up drafting boost
			slipstreamboost = Rescale(clampedkartspeed<<FRACBITS, 9<<FRACBITS, FRACUNIT, cv_mindraftspeedboostpercent.value, cv_maxdraftspeedboostpercent.value)
			slipstreamaccel = 0
			
			mo.slipstreamsoundtimer = false
		end
	end
end, MT_PLAYER)

local function varsSetup()
	local ss = "ss_"
	cv_slipstream 	= CV_FindVar(ss.."enabled")
	cv_t3chargedist = CV_FindVar(ss.."t3chargedist")
	cv_t2chargedist = CV_FindVar(ss.."t2chargedist")
	cv_maxdistance 	= CV_FindVar(ss.."maxdistance")
	cv_minimumspeed = CV_FindVar(ss.."minimumspeed")
	cv_maxangle 	= CV_FindVar(ss.."maxangle")
	cv_chargetoboost= CV_FindVar(ss.."chargetoboost")
	cv_boosttime	= CV_FindVar(ss.."boosttimer")
	cv_mindraftspeedboostpercent = CV_FindVar(ss.."mindraftspeedboostpercent")
	cv_maxdraftspeedboostpercent = CV_FindVar(ss.."maxdraftspeedboostpercent")
end

addHook("MapChange", varsSetup)
addHook("NetVars", varsSetup)

timetravel.SLIPSTREAMINTEROP_VERSION = SLIPSTREAMINTEROP_VERSION

end